from ewah.constants import EWAHConstants as EC
from ewah.operators.base import EWAHBaseOperator

from ewah.hooks.base import EWAHBaseHook as BaseHook
from airflow.utils.file import TemporaryDirectory

import os
from copy import deepcopy
from datetime import timedelta
from pymongo import MongoClient
from pymongo import ASCENDING as asc
from tempfile import NamedTemporaryFile


class EWAHMongoDBOperator(EWAHBaseOperator):

    _NAMES = ["mongo", "mongodb"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: True,
    }

    _REQUIRES_COLUMNS_DEFINITION = False

    def __init__(
        self,
        source_collection_name=None,  # defaults to target_table_name
        source_database_name=None,  # Not required if specified in connection uri
        timestamp_field=None,  # required for use with data_from and data_until
        chunking_field=None,  # defaults to timestamp_field if None
        chunking_interval=None,  # timedelta or integer
        pagination_limit=None,  # integer
        single_column_mode=False,  # If True, throw all data as json into one col
        ssl=False,
        ssl_conn_id=None,
        mongoclient_extra_args={},
        conn_style="uri",  # one of ['uri', 'credentials']
        *args,
        **kwargs
    ):

        src = source_collection_name or kwargs.get("target_table_name")
        self.source_collection_name = src
        self.source_database_name = source_database_name

        self.timestamp_field = timestamp_field
        self.ssl = ssl or bool(ssl_conn_id)
        self.ssl_conn_id = ssl_conn_id
        self.conn_style = conn_style
        self.mongoclient_extra_args = mongoclient_extra_args

        assert conn_style in ["uri", "credentials"]

        chunking_field = chunking_field or timestamp_field
        if chunking_interval:
            if not (type(chunking_interval) in [int, timedelta]):
                raise Exception("chunking_interval must be int or timedelta!")
            if not chunking_field:
                raise Exception(
                    "if chunking_interval is set, must also set "
                    "chunking_field or timestamp_field!"
                )
        self.chunking_interval = chunking_interval
        self.chunking_field = chunking_field

        if pagination_limit:
            if not type(pagination_limit) == int:
                raise Exception("pagination_limit must be an integer!")
            if not kwargs.get("primary_key_column_name"):
                raise Exception(
                    "pagination_limit must be used with " + "primary_key_column_name!"
                )

        self.pagination_limit = pagination_limit

        if single_column_mode:
            if kwargs.get("columns_definition"):
                raise Exception(
                    "single_column_mode is not compatible with " + "columns_definition!"
                )
            if not kwargs.get("extract_strategy") == EC.ES_FULL_REFRESH:
                raise Exception(
                    "single_column_mode is only compatible with "
                    + "extract_strategy = {0}!".format(EC.ES_FULL_REFRESH)
                )
            if kwargs.get("update_on_columns"):
                raise Exception(
                    "single_column_mode is not compatible with " + "update_on_columns!"
                )
            if kwargs.get("primary_key_column_name"):
                raise Exception(
                    "single_column_mode is not compatible with "
                    + "primary_key_column_name!"
                )
        self.single_column_mode = single_column_mode

        super().__init__(*args, **kwargs)

        if self.load_data_from or self.load_data_until or self.reload_data_from:
            if not timestamp_field:
                err_msg = "If using reload_data_from, load_data_from, or load_"
                err_msg += "data_until, you must also specify timestamp_field!"
                raise Exception(err_msg)

    def upload_data(self, data):
        if self.single_column_mode:
            # load all data into a single column called 'document'
            data = [{"document": x} for x in data]
            super().upload_data(data)
        else:
            # normal data loading mode
            super().upload_data(data)

    def extract_and_load_paginated(
        self,
        collection,
        filter_expressions=None,
        page_size=None,
        last_id=None,
    ):
        # adapted from https://www.codementor.io/@arpitbhayani/fast-and-efficient-pagination-in-mongodb-9095flbqr
        if page_size is None:
            # just get all data at once without pagination
            self.upload_data([i for i in collection.find(filter_expressions)])
            return

        # pagination!
        fe = deepcopy(filter_expressions or {})
        if last_id:
            # get next page
            fe.update({self.primary_key_column_name: {"$gt": last_id}})

        data = [
            i
            for i in collection.find(fe)
            .sort(self.primary_key_column_name, asc)
            .limit(page_size)
        ]

        if not len(data) == 0:
            last_id = data[-1][self.primary_key_column_name]
            self.upload_data(data)
            # call recursively!
            self.extract_and_load_paginated(
                collection=collection,
                filter_expressions=filter_expressions,
                page_size=page_size,
                last_id=last_id,
            )

    def ewah_execute(self, context):
        self.log.info("Creating filter expression...")
        base_filters = []

        # data_from and data_until filter expression
        if self.timestamp_field:
            if self.data_from:
                base_filters += [{self.timestamp_field: {"$gte": self.data_from}}]
            if self.data_until:
                base_filters += [{self.timestamp_field: {"$lt": self.data_until}}]

        self.log.info("Connecting to MongoDB Database...")
        if self.conn_style == "uri":
            conn_kwargs = {"host": self.source_conn.password}  # as uri
        else:
            conn_kwargs = {
                "host": self.source_conn.host,
                "port": self.source_conn.port,
                "tz_aware": True,
            }
            if self.source_conn.login:
                conn_kwargs["username"] = self.source_conn.login
            if self.source_conn.password:
                conn_kwargs["password"] = self.source_conn.password

        with TemporaryDirectory() as tmp_dir:
            db_name = self.source_database_name
            if self.ssl:
                conn_kwargs["tls"] = True
            with NamedTemporaryFile(dir=tmp_dir) as ssl_f:
                with NamedTemporaryFile(dir=tmp_dir) as ssl_f_pk:
                    if self.ssl_conn_id:
                        ssl_f_name = os.path.abspath(ssl_f.name)
                        ssl_f_pk_name = os.path.abspath(ssl_f_pk.name)
                        ssl_conn = BaseHook.get_connection(self.ssl_conn_id)

                        extra = ssl_conn.extra_dejson or ssl_conn.extra
                        if isinstance(extra, dict):
                            # extra is json with certificate & private key
                            certificate_text = extra.get("CERTIFICATE")
                            private_text = extra.get("PRIVATE")
                            ssl_f.write(certificate_text.encode())
                            ssl_f_pk.write(private_text.encode())
                            ssl_f.seek(0)
                            ssl_f_pk.seek(0)
                            conn_kwargs["ssl_certfile"] = ssl_f_name
                            conn_kwargs["ssl_keyfile"] = ssl_f_pk_name
                        else:
                            # extra is text with certificate & private key
                            ssl_f.write(extra.encode())
                            ssl_f.seek(0)
                            conn_kwargs["tlsCertificateKeyFile"] = ssl_f_name
                        if ssl_conn.password:
                            # works for both ways of providing SSL cert
                            conn_kwargs[
                                "tlsCertificateKeyFilePassword"
                            ] = ssl_conn.password

                    conn_kwargs.update(self.mongoclient_extra_args)
                    mdb_conn = MongoClient(**conn_kwargs)
                    database = mdb_conn.get_database(name=db_name)

        self.log.info("Getting data...")
        collection = database[self.source_collection_name]
        if self.chunking_interval:
            # get data in chunks
            if base_filters:
                match_filters = {}
                for filter in base_filters:
                    for key, value in filter.items():
                        if match_filters.get(key):
                            match_filters[key].update(value)
                        else:
                            match_filters[key] = value
                aggregation_statement = [
                    {"$match": match_filters},
                    {
                        "$group": {
                            "_id": None,
                            "min": {"$min": "$" + self.chunking_field},
                            "max": {"$max": "$" + self.chunking_field},
                        }
                    },
                ]
            else:
                aggregation_statement = [
                    {
                        "$group": {
                            "_id": None,
                            "min": {"$min": "$" + self.chunking_field},
                            "max": {"$max": "$" + self.chunking_field},
                        },
                    }
                ]
            self.log.info(
                "Getting data range using:\n{0}".format(str(aggregation_statement))
            )
            minmax = list(collection.aggregate(aggregation_statement))
            if len(minmax) == 0:
                self.log.info("There appears to be no relevant data!")
                self.upload_data([])  # Upload nothing!
                return
            if not len(minmax) == 1:
                raise Exception("Unexpected error occurred")
            minmax = minmax[0]
            current_val = minmax["min"]  # == min_val
            max_val = minmax["max"]
            self.log.info(
                "loading data from {0} to {1}.".format(
                    str(current_val),
                    str(max_val),
                )
            )
            while current_val < max_val:
                next_val = min(max_val, current_val + self.chunking_interval)
                self.log.info(
                    "Getting data from {0} to {1}...".format(
                        str(current_val),
                        str(next_val),
                    )
                )

                filter_expressions = deepcopy(base_filters)
                lt_or_lte = "$lte" if next_val == max_val else "$lt"
                filter_expressions += [
                    {
                        self.chunking_field: {
                            "$gte": current_val,
                            lt_or_lte: next_val,
                        }
                    }
                ]

                filter_expressions = {"$and": filter_expressions}
                self.log.info("Filter expression: {0}".format(str(filter_expressions)))

                self.log.info("Uploading chunk...")
                self.extract_and_load_paginated(
                    collection=collection,
                    filter_expressions=filter_expressions,
                    page_size=self.pagination_limit,
                    last_id=None,
                )

                current_val = next_val
        else:
            # Just get all data in one go
            if base_filters:
                filter_expressions = {"$and": base_filters}
            else:
                filter_expressions = None
            self.log.info("Filter expression: {0}".format(str(filter_expressions)))
            # Upload data
            self.extract_and_load_paginated(
                collection=collection,
                filter_expressions=filter_expressions,
                page_size=self.pagination_limit,
                last_id=None,
            )



"""This data loading strategy is a mix of full refresh and incremental.
It has two DAGs like the incremental strategy, but no Reset.
Both DAGs have a schedule interval.
There is a Full Refresh DAG and an Incremental DAG.

The Full Refresh DAG runs with a longer periodicity, e.g. 1 day or 1 week.
The Incremental DAG then runs multiple times within the longer periodicity,
e.g. every hour of the day. That way data is current at short time intervals,
but events that can only be captured by a Full Refresh (e.g. deletions) are
also captured.
"""

from airflow import DAG

from ewah.constants import EWAHConstants as EC
from ewah.dag_factories.dag_factory_incremental import ExtendedETS
from ewah.ewah_utils.airflow_utils import (
    etl_schema_tasks,
    datetime_utcnow_with_tz,
    EWAHSqlSensor,
)
from ewah.operators.base import EWAHBaseOperator
from ewah.hooks.base import EWAHBaseHook

from datetime import datetime, timedelta
from collections.abc import Iterable
from copy import deepcopy
from typing import Optional, Type, Callable, List, Tuple, Union

import re


def dag_factory_fullcremental(
    dag_name: str,
    dwh_engine: str,
    dwh_conn_id: str,
    airflow_conn_id: str,
    start_date: datetime,
    el_operator: Type[EWAHBaseOperator],
    operator_config: dict,
    target_schema_name: str,
    target_schema_suffix: str = "_next",
    target_database_name: Optional[str] = None,
    default_args: Optional[dict] = None,
    schedule_interval_full_refresh: timedelta = timedelta(days=1),
    schedule_interval_incremental: timedelta = timedelta(hours=1),
    end_date: Optional[datetime] = None,
    read_right_users: Optional[Union[List[str], str]] = None,
    dwh_ssh_tunnel_conn_id: Optional[str] = None,
    additional_dag_args: Optional[dict] = None,
    additional_task_args: Optional[dict] = None,
    logging_func: Optional[Callable] = None,
    **kwargs
) -> Tuple[DAG, DAG]:
    def raise_exception(msg: str) -> None:
        """Add information to error message before raising."""
        raise Exception("DAG: {0} - Error: {1}".format(dag_name, msg))

    logging_func = logging_func or print

    if kwargs:
        logging_func("unused config: {0}".format(str(kwargs)))

    additional_dag_args = additional_dag_args or {}
    additional_task_args = additional_task_args or {}

    if dwh_ssh_tunnel_conn_id and not dwh_engine == EC.DWH_ENGINE_POSTGRES:
        raise_exception("DWH tunneling only implemented for PostgreSQL DWHs!")
    if not read_right_users is None:
        if isinstance(read_right_users, str):
            read_right_users = [u.strip() for u in read_right_users.split(",")]
        if not isinstance(read_right_users, Iterable):
            raise_exception("read_right_users must be an iterable or string!")
    if not isinstance(schedule_interval_full_refresh, timedelta):
        raise_exception("schedule_interval_full_refresh must be timedelta!")
    if not isinstance(schedule_interval_incremental, timedelta):
        raise_exception("schedule_interval_incremental must be timedelta!")
    if schedule_interval_incremental >= schedule_interval_full_refresh:
        _msg = "schedule_interval_incremental must be shorter than "
        _msg += "schedule_interval_full_refresh!"
        raise_exception(_msg)

    """Calculate the datetimes and timedeltas for the two DAGs.

    Full Refresh: The start_date should be chosen such that there is always only
    one DAG execution to be executed at any given point in time.

    The Incremental DAG starts at the start date + schedule interval of the
    Full Refresh DAG, so that the Incremental executions only happen after
    the Full Refresh execution.
    """
    if not start_date.tzinfo:
        # if no timezone is given, assume UTC
        raise_exception("start_date must be timezone aware!")
    time_now = datetime_utcnow_with_tz() + schedule_interval_incremental / 2
    if start_date > time_now:
        # Start date for both is in the future
        start_date_fr = start_date
        start_date_inc = start_date
    else:
        _td = int((time_now - start_date) / schedule_interval_full_refresh) - 1
        start_date_fr = start_date + _td * schedule_interval_full_refresh
        start_date_inc = start_date_fr + schedule_interval_full_refresh

    dag_name_fr = dag_name + "_Periodic_Full_Refresh"
    dag_name_inc = dag_name + "_Intraperiod_Incremental"
    dags = (
        DAG(
            dag_name_fr,
            start_date=start_date_fr,
            end_date=end_date,
            schedule_interval=schedule_interval_full_refresh,
            catchup=True,
            max_active_runs=1,
            default_args=default_args,
            **additional_dag_args
        ),
        DAG(
            dag_name_inc,
            start_date=start_date_inc,
            end_date=end_date,
            schedule_interval=schedule_interval_incremental,
            catchup=True,
            max_active_runs=1,
            default_args=default_args,
            **additional_dag_args
        ),
    )

    kickoff_fr, final_fr = etl_schema_tasks(
        dag=dags[0],
        dwh_engine=dwh_engine,
        dwh_conn_id=dwh_conn_id,
        target_schema_name=target_schema_name,
        target_schema_suffix=target_schema_suffix,
        target_database_name=target_database_name,
        read_right_users=read_right_users,
        ssh_tunnel_conn_id=dwh_ssh_tunnel_conn_id,
        **additional_task_args
    )

    kickoff_inc, final_inc = etl_schema_tasks(
        dag=dags[1],
        dwh_engine=dwh_engine,
        dwh_conn_id=dwh_conn_id,
        target_schema_name=target_schema_name,
        target_schema_suffix=target_schema_suffix,
        target_database_name=target_database_name,
        read_right_users=read_right_users,
        ssh_tunnel_conn_id=dwh_ssh_tunnel_conn_id,
        **additional_task_args
    )

    sql_fr = """
        SELECT
             -- only run if there are no active DAGs that have to finish first
            CASE WHEN COUNT(*) = 0 THEN 1 ELSE 0 END
        FROM public.dag_run
        WHERE state = 'running'
          AND (
                (dag_id = '{0}' AND execution_date < '{1}')
            OR  (dag_id = '{2}' AND execution_date < '{3}')
          )
    """.format(
        dags[0]._dag_id,  # fr
        "{{ execution_date }}",  # no previous full refresh, please!
        dags[1]._dag_id,  # inc
        "{{ next_execution_date }}",  # no old incremental running, please!
    )

    # Sense if a previous instance runs OR if any incremental loads run
    # except incremental load of the same time, which is expected and waits
    fr_snsr = EWAHSqlSensor(
        task_id="sense_run_validity",
        conn_id=airflow_conn_id,
        sql=sql_fr,
        dag=dags[0],
        poke_interval=5 * 60,
        mode="reschedule",  # don't block a worker and pool slot
        **additional_task_args
    )

    # Sense if a previous instance is complete excepts if its the first, then
    # check for a full refresh of the same time
    inc_ets = ExtendedETS(
        task_id="sense_run_validity",
        allowed_states=["success"],
        external_dag_id=dags[1]._dag_id,
        external_task_id=final_inc.task_id,
        execution_delta=schedule_interval_incremental,
        backfill_dag_id=dags[0]._dag_id,
        backfill_external_task_id=final_fr.task_id,
        backfill_execution_delta=schedule_interval_full_refresh,
        dag=dags[1],
        poke_interval=5 * 60,
        mode="reschedule",  # don't block a worker and pool slot
        **additional_task_args
    )

    fr_snsr >> kickoff_fr
    inc_ets >> kickoff_inc

    for table in operator_config["tables"].keys():
        arg_dict_inc = deepcopy(additional_task_args)
        arg_dict_inc.update(operator_config.get("general_config", {}))
        op_conf = operator_config["tables"][table] or {}
        arg_dict_inc.update(op_conf)
        arg_dict_inc.update(
            {
                "extract_strategy": EC.ES_INCREMENTAL,
                "task_id": "extract_load_" + re.sub(r"[^a-zA-Z0-9_]", "", table),
                "dwh_engine": dwh_engine,
                "dwh_conn_id": dwh_conn_id,
                "target_table_name": op_conf.get("target_table_name", table),
                "target_schema_name": target_schema_name,
                "target_schema_suffix": target_schema_suffix,
                "target_database_name": target_database_name,
                "target_ssh_tunnel_conn_id": dwh_ssh_tunnel_conn_id,
            }
        )
        arg_dict_fr = deepcopy(arg_dict_inc)
        arg_dict_fr["extract_strategy"] = EC.ES_FULL_REFRESH

        task_fr = el_operator(dag=dags[0], **arg_dict_fr)
        task_inc = el_operator(dag=dags[1], **arg_dict_inc)

        kickoff_fr >> task_fr >> final_fr
        kickoff_inc >> task_inc >> final_inc

    return dags



from ewah.operators.base import EWAHBaseOperator
from ewah.constants import EWAHConstants as EC

from ewah.hooks.base import EWAHBaseHook as BaseHook

import stripe


class EWAHStripeOperator(EWAHBaseOperator):

    _NAMES = ["stripe"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: False,
    }

    _REQUIRES_COLUMNS_DEFINITION = False

    _UPLOAD_AFTER_ROWS = 20000

    _listable = stripe.api_resources.abstract.ListableAPIResource
    _singleton = stripe.api_resources.abstract.SingletonAPIResource

    def __init__(self, *args, resource=None, expand=None, **kwargs):
        if resource is None:
            resource = kwargs.get("target_table_name")
        _msg = "Not a valid resource: {0}!".format(resource)
        assert hasattr(stripe, resource), _msg
        resource = getattr(stripe, resource)
        # use issubclass instead of isinstance
        assert issubclass(resource, (self._listable, self._singleton)), _msg
        kwargs["primary_key_column_name"] = "id"
        super().__init__(*args, **kwargs)
        self.resource = resource
        self.expand = expand

    def ewah_execute(self, context):
        # authenticate using API key
        stripe.api_key = self.source_conn.password

        # get data
        data = []
        resource = self.resource
        if issubclass(resource, self._listable):
            all_items = resource.list(limit=100, expand=self.expand)
            for item in all_items.auto_paging_iter():
                data += [item.to_dict_recursive()]
                if len(data) > self._UPLOAD_AFTER_ROWS:
                    # intermittend data upload
                    self.upload_data(data)
                    data = []
        elif issubclass(resource, self._singleton):
            data = [resource.retrieve(expand=self.expand).to_dict_recursive()]
            # singletons have no id, but we need it as PK -> set it manually:
            data[0]["id"] = 1
            assert len(data) == 1, "Not a singleton!"
        else:
            raise Exception("Invalid resource!")

        # final data upload
        self.upload_data(data)



from ewah.hooks.base import EWAHBaseHook

import os

from tempfile import NamedTemporaryFile

import sshtunnel

# The default timeout of 10s is too little if airflow runs many EL tasks in
# parallel -> if the SSH connection attempts are done in parallel, the timeout
# hits before the connection is established.
sshtunnel.TUNNEL_TIMEOUT = 30


def start_ssh_tunnel(ssh_conn_id, remote_conn_id, custom_timeout=None):
    """
    Start the SSH Tunnel with appropriate port forwarding.

    Returns both the SSHTunnelForwarder object for later closure and an
    adjusted connection object (host and port are adjusted in accordance with
    the open tunnel)
    """
    if custom_timeout:
        previous_timeout = sshtunnel.TUNNEL_TIMEOUT
        sshtunnel.TUNNEL_TIMEOUT = custom_timeout

    ssh_conn = EWAHBaseHook.get_connection(ssh_conn_id)
    remote_conn = EWAHBaseHook.get_connection(remote_conn_id)
    kwargs = {
        "ssh_address_or_host": (ssh_conn.host, ssh_conn.port or 22),
        "remote_bind_address": (remote_conn.host, remote_conn.port),
    }
    if ssh_conn.login:
        kwargs["ssh_username"] = ssh_conn.login or None
    if ssh_conn.password:
        kwargs["ssh_password"] = ssh_conn.password

    with NamedTemporaryFile() as keyfile:
        extra = ssh_conn.extra
        extra_dejson = ssh_conn.extra_dejson
        if extra:
            if extra_dejson:
                keyfile.write(extra_dejson.get("extra__private_key", "").encode())
            else:
                keyfile.write(extra.encode())
            keyfile.flush()
            kwargs["ssh_pkey"] = os.path.abspath(keyfile.name)
        ssh_tunnel_forwarder = sshtunnel.SSHTunnelForwarder(**kwargs)
        ssh_tunnel_forwarder.start()

    remote_conn.host = "localhost"
    remote_conn.port = ssh_tunnel_forwarder.local_bind_port

    if custom_timeout:
        sshtunnel.TUNNEL_TIMEOUT = previous_timeout

    return ssh_tunnel_forwarder, remote_conn



from ewah.hooks.base import EWAHBaseHook

import requests
import time


class EWAHAircallHook(EWAHBaseHook):

    _ATTR_RELABEL = {
        "api_id": "login",
        "api_token": "password",
    }

    conn_name_attr = "aircall_conn_id"
    default_conn_name = "aircall_default"
    conn_type = "ewah_aircall"
    hook_name = "EWAH Aircall Connection"

    _RESOURCES = {
        "users": {"incremental": True},
        "teams": {},
        "calls": {"incremental": True},
        "numbers": {"incremental": True},
        "contacts": {"incremental": True},
        "tags": {},
    }
    _BASE_URL = "https://api.aircall.io/v1/{0}"

    @staticmethod
    def get_ui_field_behaviour():
        return {
            "hidden_fields": ["port", "schema", "extra", "host"],
            "relabeling": {
                "login": "Basic Auth API ID",
                "password": "Baisc Auth API Token",
            },
        }

    def get_data_in_batches(
        self,
        resource,
        data_from=None,
        data_until=None,
        batch_size=10000,
        batch_call_pause_seconds=1,
    ):
        _msg = "batch_size param must be a positive integer <= 10k "
        assert isinstance(batch_size, int), _msg
        assert batch_size > 0, _msg
        assert batch_size <= 10000, _msg

        page_size = 50  # maximum page size is 50

        auth = requests.auth.HTTPBasicAuth(
            self.conn.api_id,
            self.conn.api_token,
        )
        url = self._BASE_URL.format(resource)
        params = {
            "per_page": page_size,
        }
        if data_from:
            params["from"] = int(time.mktime(data_from.timetuple()))
        if data_until:
            params["to"] = int(time.mktime((data_until).timetuple()))

        data = []
        while url:
            time.sleep(batch_call_pause_seconds)
            request = requests.get(url, params=params, auth=auth)
            assert request.status_code == 200, request.text
            response = request.json()
            url = response.get("meta", {}).get("next_page_link")
            data += response.get(resource, [])
            if (not url) or (len(data) + page_size > batch_size):
                yield data
                data = []



from ewah.constants import EWAHConstants as EC
from ewah.operators.base import EWAHBaseOperator

import boto3


class EWAHDynamoDBOperator(EWAHBaseOperator):

    _NAMES = ["dynamodb"]

    # For incremental loading, use Kinesis Firehose to push changes to S3
    # and use S3 operator instead
    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: False,
    }

    _REQUIRES_COLUMNS_DEFINITION = False

    def __init__(
        self,
        source_table_name=None,  # defaults to target_table_name
        pagination_limit=None,  # optionally set a pagination limit
        region_name=None,  # must provide region, alternatively via connection
        filter_expression=None,
        *args,
        **kwargs
    ):

        source_table_name = source_table_name or kwargs.get("target_table_name")
        _msg = "DynamoDBOperator requires primary_key_column_name!"
        assert kwargs.get("primary_key_column_name"), _msg
        super().__init__(*args, **kwargs)
        self.source_table_name = source_table_name
        self.pagination_limit = pagination_limit
        self.region_name = region_name
        self.filter_expression = filter_expression

    def ewah_execute(self, context):
        # TODO: create / update filter_expression with data_from / data_until

        # Get credentials and connect to table
        conn = self.source_conn
        resource_kwargs = {
            "aws_access_key_id": conn.login,
            "aws_secret_access_key": conn.password,
        }
        if isinstance(conn.extra_dejson, dict):
            resource_kwargs.update(conn.extra_dejson)
        if self.region_name:
            resource_kwargs["region_name"] = self.region_name
        resource = boto3.resource("dynamodb", **resource_kwargs)
        table = resource.Table(self.source_table_name)

        # Paginate through table
        scan_kwargs = {}
        if self.pagination_limit:
            scan_kwargs["Limit"] = self.pagination_limit
        if self.filter_expression:
            scan_kwargs["FilterExpression"] = self.filter_expression
        keep_going = True
        while keep_going:
            response = table.scan(**scan_kwargs)
            self.upload_data(response.get("Items"))
            keep_going = response.get("LastEvaluatedKey") or False
            scan_kwargs["ExclusiveStartKey"] = keep_going



from ewah.operators.sql_base import EWAHSQLBaseOperator
from ewah.constants import EWAHConstants as EC

from ewah.hooks.base import EWAHBaseHook as BaseHook

import cx_Oracle


class EWAHOracleSQLOperator(EWAHSQLBaseOperator):

    _NAMES = ["oracle"]

    _SQL_BASE = """
        SELECT\n{columns}\nFROM "{schema}"."{table}"\nWHERE {where_clause}
    """
    _SQL_BASE_SELECT = "SELECT * FROM ({select_sql}) t WHERE {{0}}"
    _SQL_COLUMN_QUOTE = '"'
    _SQL_MINMAX_CHUNKS = """
        WITH base AS ({base})
        SELECT MIN({column}), MAX({column})
        FROM base
    """
    _SQL_CHUNKING_CLAUSE = """
        AND {column} >= :from_value
        AND {column} <{equal_sign} :until_value
    """
    _SQL_PARAMS = ":{0}"

    def _get_data_from_sql(
        self,
        sql,
        params={},
        return_dict=True,
    ):
        """In Oracle, params are passed to the execute() function as kwargs
        https://cx-oracle.readthedocs.io/en/latest/user_guide/bind.html
        Params are then referenced like so:
            SELECT * FROM table WHERE field1 = :val1 AND field2 = :val2
        with a params dict structured like so: {'val1': 1, 'val2': 3}
        Also see here: https://stackoverflow.com/questions/35045879/cx-oracle-how-can-i-receive-each-row-as-a-dictionary
        to understand the workaround regarding return_dict
        """

        def makeDictFactory(cursor):
            columnNames = [d[0] for d in cursor.description]

            def createRow(*args):
                return dict(zip(columnNames, args))

            return createRow

        oracle_conn = cx_Oracle.connect(
            self.source_conn.login,
            self.source_conn.password,
            "{0}:{1}/{2}".format(
                self.source_conn.host,
                self.source_conn.port,
                self.source_conn.schema,
            ),
            encoding="UTF-8",
        )
        cursor = oracle_conn.cursor()
        if sql.strip()[-1:] == ";":  # OracleSQL doesn't like semicolons
            sql = sql.strip()[:-1]
        self.log.info(
            "Executing:\n{0}\n\nWith params:\n{1}".format(
                sql,
                str(params),
            )
        )
        cursor.execute(sql, **params)
        if return_dict:
            cursor.rowfactory = makeDictFactory(cursor)
        data = cursor.fetchall()
        cursor.close()
        oracle_conn.close()
        return data



from wtforms import widgets


class EWAHTextAreaWidget(widgets.TextArea):
    """Make a variable-depth Text Area widget."""

    def __init__(self, rows=3):
        self.rows = rows
        return super().__init__()

    def __call__(self, field, **kwargs):
        kwargs["class"] = u"form-control"
        kwargs["rows"] = kwargs.get("rows", self.rows)
        if field.label:
            kwargs["placeholder"] = field.label.text
        return super().__call__(field, **kwargs)
